@@ -26,6 +26,9 @@
 #include <unistd.h>
 #include <errno.h>
 #include <regex.h>
+#if HAVE_CRIU && HAVE_DLOPEN
+#  include <criu/criu.h>
+#endif
 
 #include "crun.h"
 #include "libcrun/container.h"
@@ -41,6 +44,9 @@ enum
   OPTION_SHELL_JOB,
   OPTION_EXT_UNIX_SK,
   OPTION_FILE_LOCKS,
+  OPTION_PARENT_PATH,
+  OPTION_PRE_DUMP,
+  OPTION_MANAGE_CGROUPS_MODE,
 };
 
 static char doc[] = "OCI runtime";
@@ -55,12 +61,36 @@ static struct argp_option options[]
         { "ext-unix-sk", OPTION_EXT_UNIX_SK, 0, 0, "allow external unix sockets", 0 },
         { "shell-job", OPTION_SHELL_JOB, 0, 0, "allow shell jobs", 0 },
         { "file-locks", OPTION_FILE_LOCKS, 0, 0, "allow file locks", 0 },
+#ifdef CRIU_PRE_DUMP_SUPPORT
+        { "parent-path", OPTION_PARENT_PATH, "DIR", 0, "path for previous criu image files in pre-dump", 0 },
+        { "pre-dump", OPTION_PRE_DUMP, 0, 0, "dump container's memory information only, leave the container running after this", 0 },
+#endif
+        { "manage-cgroups-mode", OPTION_MANAGE_CGROUPS_MODE, "MODE", 0, "cgroups mode: 'soft' (default), 'ignore', 'full' and 'strict'", 0 },
         {
             0,
         } };
 
 static char args_doc[] = "checkpoint CONTAINER";
 
+int
+crun_parse_manage_cgroups_mode (char *param arg_unused)
+{
+#if HAVE_CRIU && HAVE_DLOPEN
+  if (strcmp (param, "soft") == 0)
+    return CRIU_CG_MODE_SOFT;
+  else if (strcmp (param, "ignore") == 0)
+    return CRIU_CG_MODE_IGNORE;
+  else if (strcmp (param, "full") == 0)
+    return CRIU_CG_MODE_FULL;
+  else if (strcmp (param, "strict") == 0)
+    return CRIU_CG_MODE_STRICT;
+  else
+    libcrun_fail_with_error (0, "unknown cgroup mode specified");
+#else
+  return 0;
+#endif
+}
+
 static error_t
 parse_opt (int key, char *arg arg_unused, struct argp_state *state arg_unused)
 {
@@ -77,6 +107,14 @@ parse_opt (int key, char *arg arg_unused, struct argp_state *state arg_unused)
       cr_options.work_path = argp_mandatory_argument (arg, state);
       break;
 
+    case OPTION_PARENT_PATH:
+      cr_options.parent_path = argp_mandatory_argument (arg, state);
+      break;
+
+    case OPTION_PRE_DUMP:
+      cr_options.pre_dump = true;
+      break;
+
     case OPTION_LEAVE_RUNNING:
       cr_options.leave_running = true;
       break;
@@ -97,6 +135,10 @@ parse_opt (int key, char *arg arg_unused, struct argp_state *state arg_unused)
       cr_options.file_locks = true;
       break;
 
+    case OPTION_MANAGE_CGROUPS_MODE:
+      cr_options.manage_cgroups_mode = crun_parse_manage_cgroups_mode (argp_mandatory_argument (arg, state));
+      break;
+
     default:
       return ARGP_ERR_UNKNOWN;
     }
@@ -117,6 +159,8 @@ crun_command_checkpoint (struct crun_global_arguments *global_args, int argc, ch
     0,
   };
 
+  cr_options.manage_cgroups_mode = -1;
+
   argp_parse (&run_argp, argc, argv, ARGP_IN_ORDER, &first_arg, &cr_options);
   crun_assert_n_args (argc - first_arg, 1, 2);
 
