@@ -212,9 +212,33 @@ write_blkio_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux
       uint32_t val = blkio->weight;
 
       len = sprintf (fmt_buf, "%" PRIu32, val);
-      ret = write_cgroup_file (dirfd, cgroup2 ? "io.bfq.weight" : "blkio.weight", fmt_buf, len, err);
-      if (UNLIKELY (ret < 0))
-        return ret;
+      if (! cgroup2)
+        {
+          ret = write_cgroup_file (dirfd, "blkio.weight", fmt_buf, len, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
+      else
+        {
+          ret = write_cgroup_file (dirfd, "io.bfq.weight", fmt_buf, len, err);
+          if (UNLIKELY (ret < 0))
+            {
+              if (crun_error_get_errno (err) == ENOENT)
+                {
+                  crun_error_release (err);
+
+                  /* convert linearly from [10-1000] to [1-10000] */
+                  val = 1 + (val - 10) * 9999 / 990;
+
+                  len = sprintf (fmt_buf, "%" PRIu32, val);
+
+                  ret = write_cgroup_file (dirfd, "io.weight", fmt_buf, len, err);
+                }
+
+              if (UNLIKELY (ret < 0))
+                return ret;
+            }
+        }
     }
   if (blkio->leaf_weight)
     {
@@ -417,7 +441,6 @@ write_devices_resources_v1 (int dirfd, runtime_spec_schema_defs_linux_device_cgr
     "c 5:1 rwm",
     "c 136:* rwm",
     "c 5:2 rwm",
-    "c 10:200 rwm",
     NULL
   };
 
@@ -495,7 +518,6 @@ write_devices_resources_v2_internal (int dirfd, runtime_spec_schema_defs_linux_d
     { 'c', 5, 1, "rwm" },
     { 'c', 136, -1, "rwm" },
     { 'c', 5, 2, "rwm" },
-    { 'c', 10, 200, "rwm" },
   };
 
   program = bpf_program_new (2048);
@@ -691,6 +713,40 @@ write_memory_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linu
   char fmt_buf[32];
   bool memory_limits_written = false;
 
+  if (cgroup2 && memory->check_before_update_present && memory->check_before_update)
+    {
+      cleanup_free char *swap_current = NULL;
+      cleanup_free char *current = NULL;
+      uint64_t limit = 0;
+      uint64_t val, val_swap;
+      int ret;
+
+      ret = read_all_file_at (dirfd, "memory.current", &current, NULL, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
+      ret = read_all_file_at (dirfd, "memory.swap.current", &swap_current, NULL, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
+      errno = 0;
+      val = strtoll (current, NULL, 10);
+      if (UNLIKELY (errno))
+        return crun_make_error (err, errno, "parse memory.current");
+
+      val_swap = strtoll (swap_current, NULL, 10);
+      if (UNLIKELY (errno))
+        return crun_make_error (err, errno, "parse memory.swap.current");
+
+      if (memory->limit_present && memory->limit >= 0)
+        limit = memory->limit;
+      if (memory->swap_present && memory->swap >= 0)
+        limit += memory->swap;
+
+      if (limit <= val + val_swap)
+        return crun_make_error (err, 0, "cannot set the memory limit lower than its current usage");
+    }
+
   if (memory->limit_present)
     {
       ret = write_memory (dirfd, cgroup2, memory, err);
@@ -906,14 +962,34 @@ write_cpuset_resources (int dirfd_cpuset, int cgroup2, runtime_spec_schema_confi
       ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "cpuset.cpus", cpu->cpus, strlen (cpu->cpus),
                                                  err);
       if (UNLIKELY (ret < 0))
-        return ret;
+        {
+          if (! cgroup2 && errno == ENOENT)
+            {
+              crun_error_release (err);
+
+              ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "cpus", cpu->cpus, strlen (cpu->cpus),
+                                                         err);
+            }
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
     }
   if (cpu->mems)
     {
       ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "cpuset.mems", cpu->mems, strlen (cpu->mems),
                                                  err);
       if (UNLIKELY (ret < 0))
-        return ret;
+        {
+          if (! cgroup2 && errno == ENOENT)
+            {
+              crun_error_release (err);
+
+              ret = write_file_and_check_controllers_at (cgroup2, dirfd_cpuset, "mems", cpu->mems, strlen (cpu->mems),
+                                                         err);
+            }
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
     }
   return 0;
 }
