@@ -617,6 +617,38 @@ append_resources (sd_bus_message *m,
   return 0;
 }
 
+static int
+reset_failed_unit (sd_bus *bus, const char *unit)
+{
+  int sd_err;
+  sd_bus_error error = SD_BUS_ERROR_NULL;
+  sd_bus_message *m = NULL, *reply = NULL;
+
+  sd_err = sd_bus_message_new_method_call (bus, &m, "org.freedesktop.systemd1", "/org/freedesktop/systemd1",
+                                           "org.freedesktop.systemd1.Manager", "ResetFailedUnit");
+  if (UNLIKELY (sd_err < 0))
+    goto exit;
+
+  sd_err = sd_bus_message_append (m, "s", unit);
+  if (UNLIKELY (sd_err < 0))
+    goto exit;
+
+  sd_err = sd_bus_call (bus, m, 0, &error, &reply);
+  if (UNLIKELY (sd_err < 0))
+    goto exit;
+
+  sd_err = 0;
+
+exit:
+  if (m)
+    sd_bus_message_unref (m);
+  if (reply)
+    sd_bus_message_unref (reply);
+  sd_bus_error_free (&error);
+
+  return sd_err;
+}
+
 static int
 enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resources,
                             int cgroup_mode,
@@ -629,7 +661,7 @@ enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resource
   sd_bus_message *reply = NULL;
   int sd_err, ret = 0;
   sd_bus_error error = SD_BUS_ERROR_NULL;
-  const char *object;
+  const char *object = NULL;
   struct systemd_job_removed_s job_data = {};
   int i;
   const char *boolean_opts[10];
@@ -729,6 +761,8 @@ enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resource
       goto exit;
     }
 
+  sd_err = sd_bus_message_append (m, "(sv)", "DefaultDependencies", "b", 0);
+
   for (i = 0; boolean_opts[i]; i++)
     {
       sd_err = sd_bus_message_append (m, "(sv)", boolean_opts[i], "b", 1);
@@ -760,8 +794,22 @@ enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resource
   sd_err = sd_bus_call (bus, m, 0, &error, &reply);
   if (UNLIKELY (sd_err < 0))
     {
-      ret = crun_make_error (err, sd_bus_error_get_errno (&error), "sd-bus call: %s", error.message ?: error.name);
-      goto exit;
+      if (reset_failed_unit (bus, scope) == 0)
+        {
+          sd_bus_error_free (&error);
+          if (reply)
+            sd_bus_message_unref (reply);
+
+          error = SD_BUS_ERROR_NULL;
+          reply = NULL;
+
+          sd_err = sd_bus_call (bus, m, 0, &error, &reply);
+        }
+      if (sd_err < 0)
+        {
+          ret = crun_make_error (err, sd_bus_error_get_errno (&error), "sd-bus call: %s", error.message ?: error.name);
+          goto exit;
+        }
     }
 
   sd_err = sd_bus_message_read (reply, "o", &object);
@@ -903,10 +951,39 @@ libcrun_cgroup_enter_systemd (struct libcrun_cgroup_args *args,
   return 0;
 }
 
+char *
+get_cgroup_scope_path (const char *cgroup_path, const char *scope)
+{
+  char *path_to_scope = NULL;
+  char *cur;
+
+  path_to_scope = xstrdup (cgroup_path);
+
+  cur = strchr (path_to_scope, '/');
+  while (cur)
+    {
+      char *next = strchr (cur + 1, '/');
+      if (next == NULL)
+        break;
+
+      *next = '\0';
+      if (strcmp (cur, scope) == 0)
+        return path_to_scope;
+      *next = '/';
+
+      cur = next;
+      while (*cur == '/')
+        cur++;
+    }
+
+  return path_to_scope;
+}
+
 static int
 libcrun_destroy_cgroup_systemd (struct libcrun_cgroup_status *cgroup_status,
                                 libcrun_error_t *err)
 {
+  cleanup_free char *path_to_scope = NULL;
   int mode;
   int ret;
 
@@ -922,7 +999,9 @@ libcrun_destroy_cgroup_systemd (struct libcrun_cgroup_status *cgroup_status,
   if (UNLIKELY (ret < 0))
     crun_error_release (err);
 
-  return destroy_cgroup_path (cgroup_status->path, mode, err);
+  path_to_scope = get_cgroup_scope_path (cgroup_status->path, cgroup_status->scope);
+
+  return destroy_cgroup_path (path_to_scope, mode, err);
 }
 
 static int
