@@ -20,7 +20,7 @@
 #define CONTAINER_H
 
 #include <config.h>
-#include <runtime_spec_schema_config_schema.h>
+#include <ocispec/runtime_spec_schema_config_schema.h>
 #include "error.h"
 
 enum handler_configure_phase
@@ -58,6 +58,9 @@ struct libcrun_context_s
   bool force_no_cgroup;
   bool no_pivot;
 
+  char **argv;
+  int argc;
+
   struct custom_handler_manager_s *handler_manager;
 };
 
@@ -169,6 +172,18 @@ LIBCRUN_PUBLIC int libcrun_container_update (libcrun_context_t *context, const c
 LIBCRUN_PUBLIC int libcrun_container_update_from_file (libcrun_context_t *context, const char *id, const char *file,
                                                        libcrun_error_t *err);
 
+struct libcrun_update_value_s
+{
+  const char *section;
+  const char *name;
+  bool numeric;
+  const char *value;
+};
+
+LIBCRUN_PUBLIC int libcrun_container_update_from_values (libcrun_context_t *context, const char *id,
+                                                         struct libcrun_update_value_s *values, size_t len,
+                                                         libcrun_error_t *err);
+
 LIBCRUN_PUBLIC int libcrun_container_spec (bool root, FILE *out, libcrun_error_t *err);
 
 LIBCRUN_PUBLIC int libcrun_container_pause (libcrun_context_t *context, const char *id, libcrun_error_t *err);
@@ -183,6 +198,8 @@ LIBCRUN_PUBLIC int libcrun_container_restore (libcrun_context_t *context, const
 
 LIBCRUN_PUBLIC int libcrun_container_read_pids (libcrun_context_t *context, const char *id, bool recurse, pid_t **pids, libcrun_error_t *err);
 
+LIBCRUN_PUBLIC int libcrun_write_json_containers_list (libcrun_context_t *context, FILE *out, libcrun_error_t *err);
+
 // Not part of the public API, just a method in container.c we need to access from linux.c
 void get_root_in_the_userns (runtime_spec_schema_config_schema *def, uid_t host_uid, gid_t host_gid,
                              uid_t *uid, gid_t *gid);
