@@ -24,7 +24,7 @@
 #include <errno.h>
 #include <argp.h>
 #include <sys/syscall.h>
-#include <runtime_spec_schema_config_schema.h>
+#include <ocispec/runtime_spec_schema_config_schema.h>
 #include "container.h"
 #include "status.h"
 
@@ -53,11 +53,18 @@ typedef int (*container_entrypoint_t) (void *args, char *notify_socket, int sync
 
 typedef int (*set_mounts_cb_t) (void *args, libcrun_error_t *err);
 
+struct libcrun_dirfd_s
+{
+  int *dirfd;
+  bool joined;
+};
+
 pid_t libcrun_run_linux_container (libcrun_container_t *container, container_entrypoint_t entrypoint, void *args,
-                                   int *sync_socket_out, libcrun_error_t *err);
+                                   int *sync_socket_out, struct libcrun_dirfd_s *dirfd, libcrun_error_t *err);
 int get_notify_fd (libcrun_context_t *context, libcrun_container_t *container, int *notify_socket_out,
                    libcrun_error_t *err);
-int libcrun_set_mounts (struct container_entrypoint_s *args, libcrun_container_t *container, const char *rootfs, set_mounts_cb_t cb, void *cb_data, libcrun_error_t *err);
+int libcrun_set_mounts (struct container_entrypoint_s *args, libcrun_container_t *container, const char *rootfs,
+                        set_mounts_cb_t cb, void *cb_data, libcrun_error_t *err);
 int libcrun_init_caps (libcrun_error_t *err);
 int libcrun_do_pivot_root (libcrun_container_t *container, bool no_pivot, const char *rootfs, libcrun_error_t *err);
 int libcrun_reopen_dev_null (libcrun_error_t *err);
@@ -66,20 +73,22 @@ int libcrun_set_caps (runtime_spec_schema_config_schema_process_capabilities *ca
                       int no_new_privileges, libcrun_error_t *err);
 int libcrun_set_rlimits (runtime_spec_schema_config_schema_process_rlimits_element **rlimits, size_t len,
                          libcrun_error_t *err);
-int libcrun_set_selinux_exec_label (runtime_spec_schema_config_schema_process *proc, libcrun_error_t *err);
-int libcrun_set_apparmor_profile (runtime_spec_schema_config_schema_process *proc, libcrun_error_t *err);
+int libcrun_set_selinux_label (runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
+int libcrun_set_apparmor_profile (runtime_spec_schema_config_schema_process *proc, bool now, libcrun_error_t *err);
 int libcrun_set_hostname (libcrun_container_t *container, libcrun_error_t *err);
+int libcrun_set_domainname (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_oom (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_sysctl (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_terminal (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_join_process (libcrun_container_t *container, pid_t pid_to_join, libcrun_container_status_t *status,
                           const char *cgroup, int detach, int *terminal_fd, libcrun_error_t *err);
-int libcrun_linux_container_update (libcrun_container_status_t *status, const char *content, size_t len,
-                                    libcrun_error_t *err);
+int libcrun_linux_container_update (libcrun_container_status_t *status,
+                                    runtime_spec_schema_config_linux_resources *resources, libcrun_error_t *err);
 int libcrun_create_keyring (const char *name, const char *label, libcrun_error_t *err);
 int libcrun_container_pause_linux (libcrun_container_status_t *status, libcrun_error_t *err);
 int libcrun_container_unpause_linux (libcrun_container_status_t *status, libcrun_error_t *err);
-int libcrun_container_do_bind_mount (libcrun_container_t *container, char *mount_source, char *mount_destination, char **mount_options, size_t mount_options_len, libcrun_error_t *err);
+int libcrun_container_do_bind_mount (libcrun_container_t *container, char *mount_source, char *mount_destination,
+                                     char **mount_options, size_t mount_options_len, libcrun_error_t *err);
 int libcrun_container_enter_cgroup_ns (libcrun_container_t *container, libcrun_error_t *err);
 int libcrun_set_personality (runtime_spec_schema_defs_linux_personality *p, libcrun_error_t *err);
 int libcrun_configure_network (libcrun_container_t *container, libcrun_error_t *err);
@@ -100,9 +109,10 @@ int libcrun_create_final_userns (libcrun_container_t *container, libcrun_error_t
 int libcrun_save_external_descriptors (libcrun_container_t *container, pid_t pid, libcrun_error_t *err);
 
 int libcrun_create_dev (libcrun_container_t *container, int devfd,
-                        struct device_s *device, bool binds,
+                        int srcfd, struct device_s *device, bool binds,
                         bool ensure_parent_dir, libcrun_error_t *err);
 
-int parse_idmapped_mount_option (runtime_spec_schema_config_schema *def, bool is_uids, char *option, char **out, size_t *len, libcrun_error_t *err);
+int parse_idmapped_mount_option (runtime_spec_schema_config_schema *def, bool is_uids, char *option, char **out,
+                                 size_t *len, libcrun_error_t *err);
 
 #endif
