@@ -324,11 +324,11 @@ crun_ensure_directory_at (int dirfd, const char *path, int mode, bool nofollow,
 static int
 check_fd_under_path (const char *rootfs, size_t rootfslen, int fd, const char *fdname, libcrun_error_t *err)
 {
-  int ret;
+  proc_fd_path_t fdpath;
   char link[PATH_MAX];
-  char fdpath[64];
+  int ret;
 
-  sprintf (fdpath, "/proc/self/fd/%d", fd);
+  get_proc_self_fd_path (fdpath, fd);
   ret = TEMP_FAILURE_RETRY (readlink (fdpath, link, sizeof (link)));
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "readlink `%s`", fdname);
@@ -707,6 +707,19 @@ check_running_in_user_namespace (libcrun_error_t *err)
   return ret;
 }
 
+static size_t
+get_page_size ()
+{
+  static atomic_long cached_pagesize = 0;
+  atomic_long pagesize = cached_pagesize;
+  if (pagesize == 0)
+    {
+      pagesize = (long) sysconf (_SC_PAGESIZE);
+      cached_pagesize = pagesize;
+    }
+  return (size_t) pagesize;
+}
+
 static int selinux_enabled = -1;
 static int apparmor_enabled = -1;
 
@@ -725,7 +738,7 @@ libcrun_initialize_selinux (libcrun_error_t *err)
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "open /proc/mounts");
 
-  ret = read_all_fd (fd, "/proc/mounts", &out, &len, err);
+  ret = read_all_fd_with_size_hint (fd, "/proc/mounts", &out, &len, get_page_size (), err);
   if (UNLIKELY (ret < 0))
     return ret;
 
@@ -767,7 +780,7 @@ libcrun_is_selinux_enabled (libcrun_error_t *err)
 }
 
 int
-add_selinux_mount_label (char **retlabel, const char *data, const char *label, libcrun_error_t *err arg_unused)
+add_selinux_mount_label (char **retlabel, const char *data, const char *label, const char *context_type, libcrun_error_t *err arg_unused)
 {
   int ret;
 
@@ -778,9 +791,9 @@ add_selinux_mount_label (char **retlabel, const char *data, const char *label, l
   if (label && ret)
     {
       if (data && *data)
-        xasprintf (retlabel, "%s,context=\"%s\"", data, label);
+        xasprintf (retlabel, "%s,%s=\"%s\"", data, context_type, label);
       else
-        xasprintf (retlabel, "context=\"%s\"", label);
+        xasprintf (retlabel, "%s=\"%s\"", context_type, label);
       return 0;
     }
   *retlabel = xstrdup (data);
@@ -866,21 +879,26 @@ set_apparmor_profile (const char *profile, bool now, libcrun_error_t *err)
 }
 
 int
-read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_error_t *err)
+read_all_fd_with_size_hint (int fd, const char *description, char **out, size_t *len, size_t size_hint, libcrun_error_t *err)
 {
-  int ret;
-  size_t nread, allocated;
-  off_t size = 0;
   cleanup_free char *buf = NULL;
+  size_t nread, allocated;
+  size_t pagesize = 0;
+  off_t size = 0;
+  int ret;
 
-  ret = get_file_size (fd, &size);
-  if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "error stat'ing file `%s`", description);
+  if (size_hint)
+    allocated = size_hint;
+  else
+    {
+      ret = get_file_size (fd, &size);
+      if (UNLIKELY (ret < 0))
+        return crun_make_error (err, errno, "error stat'ing file `%s`", description);
+
+      allocated = size == 0 ? 1023 : size;
+    }
 
   /* NUL terminate the buffer.  */
-  allocated = size;
-  if (size == 0)
-    allocated = 4096;
   buf = xmalloc (allocated + 1);
   nread = 0;
   while ((size && nread < (size_t) size) || size == 0)
@@ -894,9 +912,27 @@ read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_e
 
       nread += ret;
 
-      allocated += 4096;
-      buf = xrealloc (buf, allocated + 1);
+      if (nread == allocated)
+        {
+          if (size)
+            break;
+
+          if (pagesize == 0)
+            pagesize = get_page_size ();
+
+          allocated += pagesize;
+
+          buf = xrealloc (buf, allocated + 1);
+        }
     }
+  if (nread + 1 < allocated)
+    {
+      /* shrink the buffer to the used size if it was allocated a bigger block.  */
+      char *tmp = realloc (buf, nread + 1);
+      if (tmp)
+        buf = tmp;
+    }
+
   buf[nread] = '\0';
   *out = buf;
   buf = NULL;
@@ -931,7 +967,7 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
 {
   struct sockaddr_un addr = {};
   int ret;
-  char name_buf[32];
+  proc_fd_path_t name_buf;
   cleanup_close int destfd = -1;
   cleanup_close int fd = -1;
 
@@ -945,7 +981,8 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
       if (UNLIKELY (destfd < 0))
         return crun_make_error (err, errno, "error opening `%s`", path);
 
-      sprintf (name_buf, "/proc/self/fd/%d", destfd);
+      get_proc_self_fd_path (name_buf, destfd);
+
       path = name_buf;
     }
 
@@ -965,15 +1002,15 @@ int
 open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
 {
   struct sockaddr_un addr = {};
+  proc_fd_path_t name_buf;
   int ret;
-  char name_buf[32];
   cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error creating UNIX socket");
 
   if (strlen (path) >= sizeof (addr.sun_path))
     {
-      sprintf (name_buf, "/proc/self/fd/%d", fd);
+      get_proc_self_fd_path (name_buf, fd);
       path = name_buf;
     }
   strcpy (addr.sun_path, path);
@@ -1148,15 +1185,14 @@ copy_from_fd_to_fd (int src, int dst, int consume, libcrun_error_t *err)
 {
   int ret;
   ssize_t nread;
+  size_t pagesize = get_page_size ();
   do
     {
       cleanup_free char *buffer = NULL;
       ssize_t remaining;
 
-#define BUFFER_SIZE 4096
-
 #ifdef HAVE_COPY_FILE_RANGE
-      nread = copy_file_range (src, NULL, dst, NULL, BUFFER_SIZE, 0);
+      nread = copy_file_range (src, NULL, dst, NULL, pagesize, 0);
       if (nread < 0 && (errno == EINVAL || errno == EXDEV))
         goto fallback;
       if (consume && nread < 0 && errno == EAGAIN)
@@ -1170,8 +1206,8 @@ copy_from_fd_to_fd (int src, int dst, int consume, libcrun_error_t *err)
     fallback:
 #endif
 
-      buffer = xmalloc (BUFFER_SIZE);
-      nread = TEMP_FAILURE_RETRY (read (src, buffer, BUFFER_SIZE));
+      buffer = xmalloc (pagesize);
+      nread = TEMP_FAILURE_RETRY (read (src, buffer, pagesize));
       if (consume && nread < 0 && errno == EAGAIN)
         return 0;
       if (nread < 0 && errno == EIO)
@@ -1652,7 +1688,7 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
         {
           ret = fcntl (val, F_SETFD, FD_CLOEXEC);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "cannot set CLOEXEC fd for '/proc/self/fd/%s'", name);
+            return crun_make_error (err, errno, "cannot set CLOEXEC fd for `/proc/self/fd/%s`", name);
         }
     }
   return 0;
@@ -2045,14 +2081,15 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
         {
           if (errno == ENOTSUP)
             {
-              char proc_path[32];
+              proc_fd_path_t proc_path;
               cleanup_close int fd = -1;
 
               fd = openat (destdirfd, de->d_name, O_PATH | O_NOFOLLOW);
               if (UNLIKELY (fd < 0))
                 return crun_make_error (err, errno, "open `%s/%s`", destname, de->d_name);
 
-              sprintf (proc_path, "/proc/self/fd/%d", fd);
+              get_proc_self_fd_path (proc_path, fd);
+
               ret = chmod (proc_path, mode & ALLPERMS);
             }
 
