@@ -291,11 +291,17 @@ int check_running_in_user_namespace (libcrun_error_t *err);
 
 int set_selinux_label (const char *label, bool now, libcrun_error_t *err);
 
-int add_selinux_mount_label (char **ret, const char *data, const char *label, libcrun_error_t *err);
+int add_selinux_mount_label (char **ret, const char *data, const char *label, const char *context_type, libcrun_error_t *err);
 
 int set_apparmor_profile (const char *profile, bool now, libcrun_error_t *err);
 
-int read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_error_t *err);
+int read_all_fd_with_size_hint (int fd, const char *description, char **out, size_t *len, size_t hint, libcrun_error_t *err);
+
+static inline int
+read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_error_t *err)
+{
+  return read_all_fd_with_size_hint (fd, description, out, len, 0, err);
+}
 
 int read_all_file (const char *path, char **out, size_t *len, libcrun_error_t *err);
 
