@@ -73,7 +73,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case 'f':
-      config_file = crun_context.config_file = argp_mandatory_argument (arg, state);
+      config_file = argp_mandatory_argument (arg, state);
       break;
 
     case 'b':
@@ -119,12 +119,15 @@ static struct argp run_argp = { options, parse_opt, args_doc, doc, NULL, NULL, N
 int
 crun_command_run (struct crun_global_arguments *global_args, int argc, char **argv, libcrun_error_t *err)
 {
-  int first_arg, ret;
-  cleanup_free libcrun_container_t *container = NULL;
+  int first_arg = 0, ret;
+  cleanup_container libcrun_container_t *container = NULL;
   cleanup_free char *bundle_cleanup = NULL;
   cleanup_free char *config_file_cleanup = NULL;
 
   crun_context.preserve_fds = 0;
+  crun_context.listen_fds = 0;
+  /* Check if global handler is configured and pass it down to crun context */
+  crun_context.handler = global_args->handler;
 
   argp_parse (&run_argp, argc, argv, ARGP_IN_ORDER, &first_arg, &crun_context);
   crun_assert_n_args (argc - first_arg, 1, 1);
@@ -138,12 +141,13 @@ crun_command_run (struct crun_global_arguments *global_args, int argc, char **ar
           if (config_file_cleanup == NULL)
             libcrun_fail_with_error (errno, "realpath `%s` failed", config_file);
           config_file = config_file_cleanup;
-          crun_context.config_file = config_file;
         }
     }
 
   /* Make sure the bundle is an absolute path.  */
-  if (bundle)
+  if (bundle == NULL)
+    bundle = bundle_cleanup = getcwd (NULL, 0);
+  else
     {
       if (bundle[0] != '/')
         {
@@ -165,9 +169,12 @@ crun_command_run (struct crun_global_arguments *global_args, int argc, char **ar
   if (UNLIKELY (ret < 0))
     return ret;
 
-  crun_context.bundle = bundle ? bundle : ".";
+  crun_context.bundle = bundle;
   if (getenv ("LISTEN_FDS"))
-    crun_context.preserve_fds += strtoll (getenv ("LISTEN_FDS"), NULL, 10);
+    {
+      crun_context.listen_fds = strtoll (getenv ("LISTEN_FDS"), NULL, 10);
+      crun_context.preserve_fds += crun_context.listen_fds;
+    }
 
   return libcrun_container_run (&crun_context, container, 0, err);
 }
