@@ -1,4 +1,4 @@
-#!/bin/env $PYTHON
+#!/bin/env python3
 # crun - OCI runtime written in C
 #
 # Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
@@ -15,12 +15,6 @@
 # You should have received a copy of the GNU General Public License
 # along with crun.  If not, see <http://www.gnu.org/licenses/>.
 
-import time
-import json
-import subprocess
-import os
-import shutil
-import sys
 from tests_utils import *
 
 def test_hostname():
