@@ -1,4 +1,4 @@
-#!/bin/env $PYTHON
+#!/bin/env python3
 # crun - OCI runtime written in C
 #
 # Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
@@ -15,18 +15,12 @@
 # You should have received a copy of the GNU General Public License
 # along with crun.  If not, see <http://www.gnu.org/licenses/>.
 
-import time
-import json
-import subprocess
-import os
-import shutil
-import sys
 from tests_utils import *
 
 def test_limit_pid_0():
     conf = base_config()
     add_all_namespaces(conf)
-    if os.getuid() != 0:
+    if is_rootless():
         return 77
     conf['process']['args'] = ['/init', 'cat', '/dev/null']
     conf['linux']['resources'] = {"pids" : {"limit" : 0}}
@@ -37,7 +31,7 @@ def test_limit_pid_0():
 
 def test_limit_pid_n():
     conf = base_config()
-    if os.getuid() != 0:
+    if is_rootless():
         return 77
     add_all_namespaces(conf)
     conf['process']['args'] = ['/init', 'forkbomb', '20']
